/*****************************************************************************
*        jpeg.c
*
*  This module contains the code to read and write the JPEG output file
*  format.
*
*  for Persistence of Vision Raytracer
*  (C) Copyright 1995 Andreas Dilger
*---------------------------------------------------------------------------
*  NOTICE: This source code file is provided as an enhancement to POV-Ray
*  version 2.2.  This code falls under the same terms and conditions as that
*  on the regular POV-Ray source code.  There are strict rules under which you
*  are permitted to use this file.  The rules are in the file named POVLEGAL.DOC
*  which you should already have received with the POV-Ray source code.  If 
*  POVLEGAL.DOC is not available or for more info please contact Andreas Dilger,
*  the author of this code at adilger@enel.ucalgary.ca.  The POV-Ray Team is not
*  responsible for this code.
*
*  This code requires the use of libjpeg, the Independant JPEG Group's
*  JPEG library.  The latest version of this code is available at
*  ftp.uu.net as /graphics/jpeg/jpegsrc.v5a.tar.gz.  It is also available at
*  SIMTEL and it's mirror sites as /pub/msdos/graphics/jpgsrc5a.zip.  This
*  version of jpeg.c is based on libjpeg v5a, but it should be compatible with
*  earlier and later releases, since it only uses the basic functionality.
*
*****************************************************************************/

#include "frame.h"
#include "povproto.h"
#include "jpeglib.h"

static j_compress_ptr cinfo = NULL;
static j_decompress_ptr dinfo = NULL;
static JSAMPROW row_pointer[1] = { NULL };

extern int OutputQuality;

FILE_HANDLE *Get_Jpeg_File_Handle()
{
  FILE_HANDLE *handle;

  if ((handle = (FILE_HANDLE *) malloc(sizeof(FILE_HANDLE))) == NULL) 
  {
    fprintf (stderr, "Cannot allocate memory for output file handle\n");
    return(NULL);
  }

  handle->Default_File_Name_p = Default_Jpeg_File_Name;
  handle->Open_File_p = Open_Jpeg_File;
  handle->Write_Line_p = Write_Jpeg_Line;
  handle->Read_Line_p = Read_Jpeg_Line;
  handle->Read_Image_p = Read_Jpeg_Image;
  handle->Close_File_p = Close_Jpeg_File;
  handle->file = NULL;
  handle->buffer_size = 0;
  handle->buffer = NULL;
  return (handle);
}

char *Default_Jpeg_File_Name()
{
  return ("data.jpg");
}

int Open_Jpeg_File (handle, name, width, height, buffer_size, mode)
FILE_HANDLE *handle;
char *name;
int *width;
int *height;
int buffer_size;
int mode;
{
  struct jpeg_error_mgr *jerr;

  if ((jerr = (struct jpeg_error_mgr *)malloc(sizeof(struct jpeg_error_mgr)))
      == NULL)
  {
    fprintf(stderr,"Error allocating JPEG file structure\n");
    return(0);
  }

  handle->mode = mode;
  handle->filename = name;

  switch (mode) 
  {
  case READ_MODE:
    if ((handle->file = fopen (name, READ_FILE_STRING)) == NULL)
      return(0);

    if (buffer_size != 0) 
    {
      if ((handle->buffer = malloc (buffer_size)) == NULL)
        return(0);

      setvbuf (handle->file, handle->buffer, _IOFBF, buffer_size);
    }
    handle->buffer_size = buffer_size;

    if ((dinfo = (j_decompress_ptr)malloc(
                  sizeof(struct jpeg_decompress_struct))) == NULL)
    {
      fprintf(stderr,"Error allocating JPEG file structure\n");
      free(jerr);
      return(0);
    }

    dinfo->err = jpeg_std_error(jerr);

    jpeg_create_decompress(dinfo);
    jpeg_stdio_src(dinfo,handle->file);

    /* Read in header info from the file */
    jpeg_read_header(dinfo,TRUE);
    jpeg_start_decompress(dinfo);

    *width = dinfo->image_width;
    *height = dinfo->image_height;
    handle->width = *width;
    handle->height = *height;

    if ((row_pointer[0] = (JSAMPROW)malloc(3*handle->width)) == NULL)
    {
      fprintf(stderr,"Error allocating JPEG row buffer.\n");
      free(jerr);
      free(dinfo);
      return(0);
    }

    break;

  case WRITE_MODE:
    if ((handle->file = fopen (name, WRITE_FILE_STRING)) == NULL)
      return(0);

    if (buffer_size != 0) 
    {
      if ((handle->buffer = malloc (buffer_size)) == NULL)
        return(0);

      setvbuf (handle->file, handle->buffer, _IOFBF, buffer_size);
    }
    handle->buffer_size = buffer_size;

    if ((cinfo = (j_compress_ptr)malloc(
                  sizeof(struct jpeg_compress_struct))) == NULL)
    {
      fprintf(stderr,"Error allocating JPEG file structure\n");
      free(handle->buffer);
      free(jerr);
      return(0);
    }

    cinfo->err = jpeg_std_error(jerr); /* Set up JPEG error handler */

    jpeg_create_compress(cinfo);       /* Set up the compression structure */
    jpeg_stdio_dest(cinfo,handle->file);    /* Set it to write to the file */

    if (OutputQuality < 40)
      OutputQuality = 40;
    else if (OutputQuality > 100)
      OutputQuality = 100;

    jpeg_set_quality(cinfo,OutputQuality,TRUE);   /* Set the output quality */

    cinfo->restart_in_rows = 1;        /* Insert a restart marker every row */

    cinfo->image_width = *width;
    cinfo->image_height = *height;
    handle->width = *width;
    handle->height = *height;

    if ((row_pointer[0] = (JSAMPROW)malloc(3*handle->width)) == NULL)
    {
      fprintf(stderr,"Error allocating JPEG row buffer.\n");
      free(handle->buffer);
      free(cinfo);
      free(jerr);
      return(0);
    }

    cinfo->input_components = 3;    /* use 3 color components for the file    */
    cinfo->in_color_space = JCS_RGB;/* use RGB (instead of YCrCb for example) */

    jpeg_set_defaults(cinfo);       /* Use default parameters for everything  */
    jpeg_start_compress(cinfo,TRUE);/* Write header info out to JPEG file     */

    break;

  case APPEND_MODE:
    if ((handle->file = fopen (name, APPEND_FILE_STRING)) == NULL)
      return(0);

    if (buffer_size != 0) 
    {
      if ((handle->buffer = malloc (buffer_size)) == NULL)
        return(0);

      setvbuf (handle->file, handle->buffer, _IOFBF, buffer_size);
    }

    break;
  }
  return(1);
}

void Write_Jpeg_Line (handle, line_data, line_number)
FILE_HANDLE *handle;
COLOUR *line_data;
int line_number;
{
  register int col, j;

/* We must copy all the values because JPEG expects RGBRGB, but POV-Ray
 * stores RGB components in separate arrays */
  for (col = j = 0; col < handle->width; col++, j += 3) 
  {
    row_pointer[0][j]   = (int)floor(line_data[col].Red   * 255.0);
    row_pointer[0][j+1] = (int)floor(line_data[col].Green * 255.0);
    row_pointer[0][j+2] = (int)floor(line_data[col].Blue  * 255.0);
  }

  /* Write out a scanline */
  jpeg_write_scanlines(cinfo,row_pointer,1);

  /* Close and reopen file for integrity in case we crash.  Luckily, since
   * freopen will always give us back the original stream, we are safe,
   * since we have no way of shanging the original handle in the jpeg
   * structure.  This isn't 100% safety, since libjpeg buffers in blocks
   * of 8 scanlines so that it can do the DCT compression, so if it is
   * interrupted during the output of these 8 scanlines, the restart
   * feature will not work. */
  if (handle->buffer_size == 0) 
  {
    fflush(handle->file);
    handle->file = freopen(handle->filename, APPEND_FILE_STRING,
      handle->file);
  }

  line_number = cinfo->next_scanline;

  /* If we have written the last line, clean up */
  if (line_number == handle->height)
  {
    jpeg_finish_compress(cinfo);
  }
}

int Read_Jpeg_Line (handle, line_data, line_number)
FILE_HANDLE *handle;
COLOUR *line_data;
int *line_number;
{
  register int col, j;

  /* Try to read in a scanline */
  if (jpeg_read_scanlines(dinfo,row_pointer,1) != 1)
  {
     return(0);
  }

/* We must copy all the values because JPEG expects RGBRGB, but POV-Ray
 * stores RGB components in separate arrays
 */
  for (col = j = 0; col < handle->width; col++, j += 3) 
  {
    line_data[col].Blue  = (DBL)row_pointer[0][j]   / 255.0;
    line_data[col].Green = (DBL)row_pointer[0][j+1] / 255.0;
    line_data[col].Red   = (DBL)row_pointer[0][j+2] / 255.0;
  }

  *line_number = dinfo->output_scanline;

  if (*line_number == handle->height)
    jpeg_finish_decompress(dinfo);

  return(1);
}

void Close_Jpeg_File (handle)
FILE_HANDLE *handle;
{
  /* Free the memory used for the JPEG (de)compression process */
  if (cinfo)
  {
    if (cinfo->err)
      free(cinfo->err);
    jpeg_destroy_compress(cinfo);
  }
  else
  {
    if (dinfo->err)
      free(dinfo->err);
    jpeg_destroy_decompress(dinfo);
  }

  if (handle->file)
    fclose (handle->file);
  if (handle->buffer != NULL)
    free (handle->buffer);

  free(handle);
}

/* Reads a Jpeg image into an RGB image buffer. */
void
Read_Jpeg_Image(Image, name)
IMAGE *Image;
char *name;
{
  FILE *filep;
  IMAGE_LINE *line_data;
  struct jpeg_decompress_struct jinfo;
  struct jpeg_error_mgr jerr;
  unsigned int width, height, k;
  register int row, col, j;

  /* Start by trying to open the file */
  if ((filep = Locate_File(name, READ_FILE_STRING)) == NULL) 
  {
    fprintf (stderr, "Cannot open Jpeg file %s\n", name);
    close_all();
    exit(1);
  }

  jinfo.err = jpeg_std_error(&jerr);
  jpeg_stdio_src(&jinfo,filep);
  jpeg_read_header(&jinfo,TRUE);
  jpeg_start_decompress(&jinfo);

  width  = jinfo.image_width;
  height = jinfo.image_height;

  Image->iwidth  = width;
  Image->iheight = height;
  Image->width   = (DBL)width;
  Image->height  = (DBL)height;
  Image->Colour_Map_Size = 0;
  Image->Colour_Map = NULL;

  /* Allocate array for pointers to RGB row buffers */
  if ((Image->data.rgb_lines = (struct Image_Line *)
       malloc(height * sizeof(struct Image_Line))) == NULL) 
  {
    fprintf (stderr, "Cannot allocate memory for image: %s\n", name);
    close_all(); 
    exit(1);
  }

  k = width * sizeof(unsigned char);

  /* Allocate JPEG input row buffer */
  if ((row_pointer[0] = (JSAMPROW)malloc(3*k)) == NULL)
  {
    fprintf (stderr, "Cannot allocate memory for image: %s\n", name);
    close_all(); 
    exit(1);
  }

  /* Simple raster order JPEG image file, read in all of the pixels */
  for (row = 0; row < height; row++) 
  {
    /* Allocate RGB row buffers */
    line_data = &Image->data.rgb_lines[row];
    line_data->red   = (unsigned char *)malloc(k);
    line_data->green = (unsigned char *)malloc(k);
    line_data->blue  = (unsigned char *)malloc(k);
    if (line_data->red == NULL || line_data->green == NULL ||
        line_data->blue == NULL) 
    {
      fprintf(stderr, "Cannot allocate memory for image: %s\n", name);
      close_all();
      exit(1);
    }

    /* Read in scanline */
    if (jpeg_read_scanlines(&jinfo,row_pointer,1) != 1)
    {
      fprintf(stderr, "Premature EOF in image file: %s\n", name);
      close_all();
      exit(1);
    }

/* We must copy all the values because JPEG expects RGBRGB, but POV-Ray
 * stores RGB components in separate arrays */
    for (col = j = 0; col < width; col++, j += 3)
    {
      line_data->red[col]   = (unsigned char)row_pointer[0][j];
      line_data->green[col] = (unsigned char)row_pointer[0][j+1];
      line_data->blue[col]  = (unsigned char)row_pointer[0][j+2];
    }
  }

  /* Clean up the rest of the JPEG memory and such */
  free(row_pointer[0]);

  jpeg_finish_decompress(&jinfo);
  jpeg_destroy_decompress(&jinfo);
  fclose(filep);
}
